/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "uhub.h"


int hub_handle_message(struct hub_info* hub, struct user* u, const char* line, size_t length)
{
	int ret = 0;
	
	struct adc_message* cmd = adc_msg_parse(u, line, length);
	if (cmd)
	{
		switch (cmd->cmd)
		{
			case ADC_CMD_HSUP: ret = hub_handle_support(hub, u, cmd); break;
			case ADC_CMD_HPAS: ret = hub_handle_password(hub, u, cmd); break;
			case ADC_CMD_BINF: ret = hub_handle_info(hub, u, cmd); break;
			case ADC_CMD_HCHK: ret = hub_handle_autocheck(hub, u, cmd); break;
			
			/* These below aren't allowed in chat only hubs */
			case ADC_CMD_BSCH:
			case ADC_CMD_DSCH:
			case ADC_CMD_DRES:
			case ADC_CMD_ESCH:
			case ADC_CMD_FSCH:
			case ADC_CMD_DRCM:
			case ADC_CMD_DCTM:
				if (hub->config->chat_only && u->credentials < cred_operator)
				{
					break;
				}
				
			case ADC_CMD_EMSG:
			case ADC_CMD_DMSG:
			case ADC_CMD_BMSG:
			case ADC_CMD_FMSG:
			
			default:
				if (u->state == state_normal)
				{
					ret = route_message(hub, u, cmd);
				}
				else
				{
					ret = -1;
				}
				break;
		}
		adc_msg_free(cmd);
	}
	else
	{
		if (u->state != state_normal)
		{
			ret = -1;
		}
	}
	
	return ret;
}


int hub_handle_support(struct hub_info* hub, struct user* u, struct adc_message* cmd)
{
	int ret = 0;
	int index = 0;
	int ok = 1;
	char* arg = adc_msg_get_argument(cmd, index);
	struct timeval timeout = { TIMEOUT_HANDSHAKE, 0 };

	if (hub->status == hub_status_disabled && u->state == state_none)
	{
		on_login_failure(hub, u, hub->command_error_hub_disabled, "Hub is disabled");
		return -1;
	}
	
	while (arg)
	{
	
		if (strlen(arg) == 6)
		{
			fourcc_t fourcc = FOURCC(arg[2], arg[3], arg[4], arg[5]);
			if (strncmp(arg, "AD", 2) == 0)
			{
				user_support_add(u, fourcc);
			}
			else if (strncmp(arg, "RM", 2) == 0)
			{
				user_support_remove(u, fourcc);
			}
			else
			{
				ok = 0;
			}
		}
		else
		{
			ok = 0;
		}
		
		index++;
		hub_free(arg);
		arg = adc_msg_get_argument(cmd, index);
	}
	
	if (u->state == state_none)
	{
	
		if (index == 0) ok = 0; /* Need to support *SOMETHING*, at least BASE */
	
		if (ok)
		{
			hub_send_handshake(hub, u);
			event_add(&u->ev_read, &timeout);
		}
		else
		{
			/* disconnect user. Do not send crap during initial handshake! */
			user_disconnect(u, quit_logon_error);
			ret = -1;
		}
	}
	
	return ret;
}


int hub_handle_password(struct hub_info* hub, struct user* u, struct adc_message* cmd)
{
	char* password = adc_msg_get_argument(cmd, 0);
	int verification = password_verify(u, password);
	int ret = 0;

	switch (verification)
	{
		case user_auth_regular:
			route_message(hub, u, u->info);
			on_login_success(hub, u, cred_user);
			break;
			
		case user_auth_operator:
			route_message(hub, u, u->info);
			on_login_success(hub, u, cred_operator);
			break;
		
		case user_auth_admin:
			route_message(hub, u, u->info);
			on_login_success(hub, u, cred_admin);
			break;
		
		case user_banned:
			on_login_failure(hub, u, hub->command_ban_perm, "Banned");
			ret = -1;
			break;
			
		case user_not_allowed:
			on_login_failure(hub, u, hub->command_error_nick_invalid, "Not allowed");
			ret = -1;
			break;
			
		case user_not_found:
			on_login_failure(hub, u, hub->command_error_nick_invalid, "Not found");
			ret = -1;
			break;
			
		case user_invalid:
		default:
			on_login_failure(hub, u, hub->command_password_wrong, "Password incorrect");
			ret = -1;
			break;
	}
	
	hub_free(password);
	return ret;
}


int on_kick(struct user* u, struct adc_message* cmd)
{
	hub_log(log_error, "on_kick() not implemented");
	return -1;
}

int hub_handle_autocheck(struct hub_info* hub, struct user* u, struct adc_message* cmd)
{
	char* port_str = adc_msg_get_argument(cmd, 0);
	char* token    = adc_msg_get_argument(cmd, 1);
	int port = 0;
	
	if (!port_str || !token || strlen(token) != 4)
	{
		hub_free(port_str);
		hub_free(token);
		return -1;
	}
	
	port = uhub_atoi(port_str);
	
	if (port == 0 || port > 65535)
	{
		hub_free(port_str);
		hub_free(token);
		return -1;
	}
	
	hub_send_autocheck(hub, u, port, token);
	
	hub_free(port_str);
	hub_free(token);
	
	return 0;
}


void hub_send_autocheck(struct hub_info* hub, struct user* u, uint16_t port, const char* token)
{
	
}

void update_user_info(struct user* u, struct adc_message* cmd)
{
	char prefix[2];
	char* start;
	char* end;
	char* argument;
	size_t length;
	struct adc_message* cmd_new = adc_msg_copy(u->info);
	if (!cmd_new) return;
	
	/* hub_log(log_trace, "update_user_info(): %s. New stuff: '%s'", u->nick, cmd->cache); */
	
	adc_msg_unterminate_command(cmd);
	
	start = strchr(&cmd->cache[cmd->arg_offset-1], ' ');
	while (start && strlen(start) > 2)
	{
		prefix[0] = start[1];
		prefix[1] = start[2];
		start = &start[3];
		
		end = strchr(start, ' ');
		if (end == NULL)
			length = strlen(start);
		else
			length = &end[0] - &start[0];
		
		argument = hub_malloc(length + 1);
		if (argument == NULL)
		{
			adc_msg_free(cmd_new);
			cmd_new = 0;
			break;
		}
			
		memcpy(argument, start, length);
		argument[length] = 0;
		
		adc_msg_replace_named_argument(cmd_new, prefix, argument);
		hub_free(argument);

		if (!end) break;
		start = strchr(end, ' ');
	}
	
	if (cmd_new)
	{
		user_set_info(u, cmd_new);
	}
}


static int set_feature_cast_supports(struct user* u, struct adc_message* cmd)
{
	char *it, *tmp;
	
	if (adc_msg_has_named_argument(cmd, "SU"))
	{
		tmp = adc_msg_get_named_argument(cmd, "SU");
	
		if (u->feature_cast)
		{
			list_clear(u->feature_cast, &hub_free);
			list_destroy(u->feature_cast);
			u->feature_cast = 0;
		}
		u->feature_cast = list_create();
		
		it = tmp;
		while (strlen(it) > 4)
		{
			list_append(u->feature_cast, hub_strndup(it, 4));
			it = &it[5];
		}
		
		if (strlen(it) > 0)
		{
			list_append(u->feature_cast, hub_strdup(it));
		}
		
		hub_free(tmp);
	}
	return 0;
}


int hub_handle_info(struct hub_info* hub, struct user* u, struct adc_message* cmd_)
{
	char* tmp;
	char* tmp2;
	int want_ipv4, want_ipv6;
	struct access_info* access = 0;
	int need_auth = 0;
	
	/* Perform deep copy of command, since we need to reference count the info */
	struct adc_message* cmd = (struct adc_message*) hub_malloc(sizeof(struct adc_message));
	if (!cmd) return -1;
	memcpy(cmd, cmd_, sizeof(struct adc_message));
	cmd->cache = hub_malloc(cmd_->length + 2);
	memcpy(cmd->cache, cmd_->cache, cmd_->length);
	cmd->cache[cmd->length] = 0;
	
	adc_msg_remove_named_argument(cmd, "CT"); /* Client type flag (obsoletes BO, RG, OP, HU) */
	adc_msg_remove_named_argument(cmd, "BO"); /* Obsolete: bot flag (CT1) */
	adc_msg_remove_named_argument(cmd, "RG"); /* Obsolete: registered user flag (CT2) */
	adc_msg_remove_named_argument(cmd, "OP"); /* Obsolete: operator flag (CT3) */
	adc_msg_remove_named_argument(cmd, "HU"); /* Obsolete: hub flag (CT5) */
	adc_msg_remove_named_argument(cmd, "HI"); /* Obsolete: hidden user flag */
	adc_msg_remove_named_argument(cmd, "TO"); /* Client to client token */
	
	set_feature_cast_supports(u, cmd);
	
	if (u->state == state_protocol)
	{
		assert(u->info == NULL);
		
		if (adc_msg_has_named_argument(cmd, "NI"))
		{
			tmp = adc_msg_get_named_argument(cmd, "NI");
			memcpy(u->nick, tmp, MIN(strlen(tmp), MAX_NICK_LEN));
			hub_free(tmp);
		}
		
		if (adc_msg_has_named_argument(cmd, "ID"))
		{
			tmp = adc_msg_get_named_argument(cmd, "ID");
			memcpy(u->cid, tmp, MIN(strlen(tmp), MAX_CID_LEN));
			hub_free(tmp);
		}
		
		if (adc_msg_has_named_argument(cmd, "VE"))
		{
			tmp = adc_msg_get_named_argument(cmd, "VE");
			if (tmp)
			{
				tmp2 = adc_msg_unescape(tmp);
				if (tmp2)
				{
					memcpy(u->user_agent, tmp2, MIN(strlen(tmp2), MAX_UA_LEN));
					hub_free(tmp2);
				}
			}
			else
			{
				memcpy(u->user_agent, "unknown", 7);
			}
			hub_free(tmp);
		}
		else
		{
			memcpy(u->user_agent, "unknown", 7);
		}
		
		if (!hub_handle_info_check_nick(hub, u, cmd))
		{
			adc_msg_free(cmd);
			return -1;
		}
		
		if (!hub_handle_info_check_cid(hub, u, cmd))
		{
			on_login_failure(hub, u, NULL, "cid/pid invalid");
			adc_msg_free(cmd);
			return -1;
		}
		
		adc_msg_remove_named_argument(cmd, "PD");

		/* Set initial info */
		user_set_info(u, cmd);
		
		/* Strip IP addresses */
		adc_msg_remove_named_argument(cmd, "U6");
		want_ipv6 = adc_msg_remove_named_argument(cmd, "I6");
		adc_msg_remove_named_argument(cmd, "U4");
		want_ipv4 = adc_msg_remove_named_argument(cmd, "I4");
		
		/* Add correct/verified IP addresses instead (if requested/stripped) */
		tmp = (char*) net_get_peer_address(u->sd);
		if (tmp)
		{
			if (want_ipv4 && strchr(tmp, '.'))
			{
				adc_msg_add_named_argument(cmd, "I4", tmp);
			}
			else if (want_ipv6)
			{
				adc_msg_add_named_argument(cmd, "I6", tmp);
			}
		}
		
		tmp = adc_msg_get_named_argument(cmd, "NI");
		if (!tmp)
		{
			on_login_failure(hub, u, NULL, "Out of memory");
			adc_msg_free(cmd);
			return -1;
		}
		
		access = acl_get_access_info_by_name(hub->acl, tmp);
		hub_free(tmp);
		
		/*
		 * Check for any access rules associated with the nickname.
		 */
		if (access)
		{
			switch (access->status)
			{
				case user_banned:
					on_login_failure(hub, u, hub->command_ban_perm, "Banned by nickname");
					adc_msg_free(cmd);
					return -1;
					break;
					
				case user_not_allowed:
					on_login_failure(hub, u, hub->command_error_nick_invalid, "Nickname is not allowed");
					adc_msg_free(cmd);
					return -1;
					break;
					
				case user_not_found:
					u->credentials = cred_guest;
					if (hub->config->registered_users_only)
					{
						on_login_failure(hub, u, hub->command_registered_only, "Registered users only");
						adc_msg_free(cmd);
						return -1;
					}
					adc_msg_add_named_argument(cmd, "CT", "1");
					break;
					
				case user_auth_regular:
					u->credentials = cred_user;
					adc_msg_add_named_argument(cmd, "CT", "2");
					need_auth = 1;
					break;
					
				case user_auth_operator:
					u->credentials = cred_operator;
					adc_msg_add_named_argument(cmd, "CT", "3");
					need_auth = 1;
					break;
					
				case user_auth_admin:
					u->credentials = cred_admin;
					adc_msg_add_named_argument(cmd, "CT", "4");
					need_auth = 1;
					break;
			}
		}
		
		tmp = adc_msg_get_named_argument(cmd, "ID");
		if (!tmp)
		{
			on_login_failure(hub, u, NULL, "Out of memory");
			adc_msg_free(cmd);
			return -1;
		}
		
		/*
		 * Check if CID is banned.
		 */
		access = acl_get_access_info_by_cid(hub->acl, tmp);
		hub_free(tmp);
		if (access && access->status == user_banned)
		{
			on_login_failure(hub, u, hub->command_ban_perm, "Banned by CID");
			adc_msg_free(cmd);
			return -1;
		}
		
		/*
		 * If hub is full, don't let users in.
		 * We still allows operators and admins to enter.
		 */
		if (list_size(hub->user_list) >= hub->config->max_users && u->credentials < user_auth_operator)
		{
			on_login_failure(hub, u, hub->command_error_hub_full, "Hub full");
			adc_msg_free(cmd);
			return -1;
		}
		
		/*
		 * If user needs to be authenticated (using password),
		 * send the challenge.
		 *
		 * Otherwise, welcome the user.
		 */
		if (need_auth)
		{
			hub_send_password_challenge(hub, u);
		}
		else
		{
			route_message(hub, u, cmd);
			on_login_success(hub, u, cred_guest);
		}
	}
	else
	{
		/* These must not be allowed updated */
		adc_msg_remove_named_argument(cmd, "PD");
		adc_msg_remove_named_argument(cmd, "ID");
		adc_msg_remove_named_argument(cmd, "U6");
		adc_msg_remove_named_argument(cmd, "I6");
		adc_msg_remove_named_argument(cmd, "U4");
		adc_msg_remove_named_argument(cmd, "I4");
		
		
#if ALLOW_NICK_CHANGES
		/*
		 * If the nick is not accepted, do not relay it.
		 * Otherwise, the nickname will be updated.
		 */
		if (adc_msg_has_named_argument(cmd, "NI")) {
			if (!hub_handle_info_check_nick(hub, u, cmd))
			{
				adc_msg_remove_named_argument(cmd, "NI");
			}
		}
#else
		adc_msg_remove_named_argument(cmd, "NI");
#endif
		
		update_user_info(u, cmd);
		route_message(hub, u, cmd);
	}
	
	return 0;
}


void hub_send_support(struct hub_info* hub, struct user* u)
{
	route_to_user(u, hub->command_support);
	u->state = state_protocol;
}


void hub_send_sid(struct hub_info* hub, struct user* u)
{
	struct adc_message* command;
	
	command = adc_msg_create("ISID");
	adc_msg_add_argument(command, (const char*) sid_to_string(u->sid));
	
	route_to_user(u, command);
	adc_msg_free(command);
	
	u->state = state_protocol;
}


void hub_send_hubinfo(struct hub_info* hub, struct user* u)
{
	route_to_user(u, hub->command_info);
	
	if (hub->config->show_banner)
		route_to_user(u, hub->command_banner);
	
	u->state = state_protocol;
}


void hub_send_handshake(struct hub_info* hub, struct user* u)
{
	hub_send_support(hub, u);
	hub_send_sid(hub, u);
	hub_send_hubinfo(hub, u);
}


void hub_send_motd(struct hub_info* hub, struct user* u)
{
	if (hub->command_motd)
	{
		route_to_user(u, hub->command_motd);
	}
}


void hub_send_password_challenge(struct hub_info* hub, struct user* u)
{
	struct adc_message* igpa;
	igpa = adc_msg_create("IGPA");
	adc_msg_add_argument(igpa, password_generate_challenge(u));
	u->state = state_verify;
	route_to_user(u, igpa);
	adc_msg_free(igpa);
}


struct hub_info* hub_start_service(struct hub_config* config)
{
	struct hub_info* hub = 0;
	int server_tcp, server_udp, ret, ipv6_supported;
	struct sockaddr_storage addr;
	socklen_t sockaddr_size;
	
	hub = hub_malloc_zero(sizeof(struct hub_info));
	if (!hub)
	{
		hub_log(log_fatal, "Unable to allocate memory for hub");
		return 0;
	}
	
	ipv6_supported = net_is_ipv6_supported();
	
	if (ipv6_supported)
		hub_log(log_debug, "IPv6 supported and enabled.");
	else
		hub_log(log_debug, "IPv6 is not supported.");
	
	hub_log(log_info, "Starting server, listening on %s:%d", (ipv6_supported ? SERVER_ADDR_IPV6 : SERVER_ADDR_IPV4), config->server_port);

	server_tcp = net_socket_create((ipv6_supported ? PF_INET6 : PF_INET), SOCK_STREAM, IPPROTO_TCP);
	server_udp = net_socket_create((ipv6_supported ? PF_INET6 : PF_INET), SOCK_DGRAM,  IPPROTO_UDP);
	if (server_tcp == -1 || server_udp == -1)
	{
		hub_free(hub);
		if (server_tcp != -1) { net_close(server_tcp); }
		if (server_udp != -1) { net_close(server_udp); }
		return 0;
	}
	
	
	if (ip_convert_address(config->server_bind_addr, config->server_port, (struct sockaddr*) &addr, &sockaddr_size) == -1)
	{
		hub_free(hub);
		net_close(server_tcp);
		net_close(server_udp);
		return 0;
	}
	
	addr.ss_family = ipv6_supported ? AF_INET6 : AF_INET;

	ret = net_set_reuseaddress(server_tcp, 1);
	if (ret == -1)
	{
		hub_free(hub);
		net_close(server_tcp);
		net_close(server_udp);
		return 0;
	}
	
	ret = net_set_reuseaddress(server_udp, 1);
	if (ret == -1)
	{
		hub_free(hub);
		net_close(server_tcp);
		net_close(server_udp);
		return 0;
	}

	
	ret = net_set_nonblocking(server_tcp, 1);
	if (ret == -1)
	{
		hub_free(hub);
		net_close(server_tcp);
		net_close(server_udp);
		return 0;
	}
	
	ret = net_set_nonblocking(server_udp, 1);
	if (ret == -1)
	{
		hub_free(hub);
		net_close(server_tcp);
		net_close(server_udp);
		return 0;
	}
	
	
	ret = net_bind(server_tcp, (struct sockaddr*) &addr, sockaddr_size);
	if (ret == -1)
	{
		hub_log(log_fatal, "hub_start_service(): Unable to bind to TCP local address. errno=%d, str=%s", net_error(), net_error_string(net_error()));
		hub_free(hub);
		net_close(server_tcp);
		net_close(server_udp);
		return 0;
	}

	ret = net_bind(server_udp, (struct sockaddr*) &addr, sockaddr_size);
	if (ret == -1)
	{
		hub_log(log_fatal, "hub_start_service(): Unable to bind to UDP local address. errno=%d, str=%s", net_error(), net_error_string(net_error()));
		hub_free(hub);
		net_close(server_tcp);
		net_close(server_udp);
		return 0;
	}

	ret = net_listen(server_tcp, SERVER_BACKLOG);
	if (ret == -1)
	{
		hub_log(log_fatal, "hub_start_service(): Unable to listen to socket");
		hub_free(hub);
		net_close(server_tcp);
		net_close(server_udp);
		return 0;
	}
	
	hub->fd_tcp = server_tcp;
	hub->fd_udp = server_udp;
	hub->config = config;
	hub->user_list = NULL;
	hub->free_sid = 1;
	
	if (user_manager_init(hub) == -1)
	{
		hub_free(hub);
		net_close(server_tcp);
		net_close(server_udp);
		return 0;
	}
	
	event_set(&hub->ev_accept, hub->fd_tcp, EV_READ | EV_PERSIST, net_on_accept, hub);
	if (event_add(&hub->ev_accept, NULL) == -1)
	{
		user_manager_shutdown(hub);
		hub_free(hub);
		net_close(server_tcp);
		net_close(server_udp);
		return 0;
	}
	
	event_set(&hub->ev_datagram, hub->fd_udp, EV_READ | EV_PERSIST, net_on_packet, hub);
	if (event_add(&hub->ev_datagram, NULL) == -1)
	{
		user_manager_shutdown(hub);
		hub_free(hub);
		net_close(server_tcp);
		net_close(server_udp);
		return 0;
	}
	
	hub->status = hub_status_running;
	
	return hub;
}


void hub_shutdown_service(struct hub_info* hub)
{
	event_del(&hub->ev_accept);
	event_del(&hub->ev_datagram);
	net_close(hub->fd_tcp);
	net_close(hub->fd_udp);
	user_manager_shutdown(hub);
	hub->status = hub_status_stopped;
	hub_free(hub);
	hub = 0;
}


void hub_set_variables(struct hub_info* hub, struct acl_handle* acl)
{
	int fd, ret;
	char buf[MAX_RECV_BUF];
	char* tmp;
	
	hub->acl = acl;
	
	hub->command_info = adc_msg_create("IINF CT5 VE" PRODUCT "/" VERSION "");
	
	tmp = adc_msg_escape(hub->config->hub_name);
	adc_msg_add_named_argument(hub->command_info, "NI", tmp);
	hub_free(tmp);
	
	tmp = adc_msg_escape(hub->config->hub_description);
	adc_msg_add_named_argument(hub->command_info, "DE", tmp);
	hub_free(tmp);
	
	/* (Re-)read the message of the day */
	hub->command_motd = 0;
	fd = open(hub->config->file_motd, 0);
	if (fd != -1)
	{
		ret = read(fd, buf, MAX_RECV_BUF);
		if (ret > 0)
		{
			buf[ret] = 0;
			hub->command_motd = adc_msg_create("IMSG");
			tmp = adc_msg_escape(buf);
 			adc_msg_add_argument(hub->command_motd, tmp);
			hub_free(tmp);
		}
		else
		{
			
		}
		close(fd);
	}
	
	hub->command_support            = adc_msg_create("ISUP " ADC_PROTO_SUPPORT "");
	hub->command_error_nick_missing = adc_msg_create("ISTA 243 No\\snick FLNI");
	hub->command_error_nick_invalid = adc_msg_create("ISTA 221 Nick\\sinvalid");
	hub->command_error_nick_taken   = adc_msg_create("ISTA 221 Nick\\sinvalid");
	hub->command_error_cid_missing  = adc_msg_create("ISTA 243 CID\\smissing FLID");
	hub->command_error_cid_invalid  = adc_msg_create("ISTA 227 CID\\sinvalid");
	hub->command_error_cid_taken    = adc_msg_create("ISTA 224 CID\\staken");
	hub->command_error_pid_missing  = adc_msg_create("ISTA 243 PID\\smissing FLPD");
	hub->command_ban_perm           = adc_msg_create("ISTA 231 Banned");
	hub->command_ban_temp           = adc_msg_create("ISTA 232 Banned TL600"); /* FIXME: 5 minutes temp ban?! */
	hub->command_registered_only    = adc_msg_create("ISTA 226 Registered\\susers\\sonly");
	hub->command_password_wrong     = adc_msg_create("ISTA 223 Invalid\\spassword");
	hub->command_error_hub_full     = adc_msg_create("ISTA 211 Hub\\sis\\sfull");
	hub->command_error_hub_disabled = adc_msg_create("ISTA 212 Hub\\sis\\sdisabled");
	hub->command_banner             = adc_msg_create("IMSG Powered\\sby\\s" PRODUCT "/" VERSION "");
	
	/* Special command, just a linebreak */
	hub->command_ping               = adc_msg_create("IPNG");
	hub->command_ping->cache[0]     = '\n';
	hub->command_ping->cache[1]     = 0;
	hub->command_ping->length       = 1;
	hub->command_ping->priority     = 1;
	hub->command_ping->arg_offset   = 0;
	
	hub->tm_started = time(0);

	hub->status = (hub->config->hub_enabled ? hub_status_running : hub_status_disabled);
}


void hub_free_variables(struct hub_info* hub)
{
	adc_msg_free(hub->command_info);
	adc_msg_free(hub->command_banner);
	
	if (hub->command_motd)
		adc_msg_free(hub->command_motd);
	
	adc_msg_free(hub->command_support);
	adc_msg_free(hub->command_error_nick_missing);
	adc_msg_free(hub->command_error_nick_invalid);
	adc_msg_free(hub->command_error_nick_taken);
	adc_msg_free(hub->command_error_cid_missing);
	adc_msg_free(hub->command_error_cid_invalid);
	adc_msg_free(hub->command_error_cid_taken);
	adc_msg_free(hub->command_error_pid_missing);
	adc_msg_free(hub->command_ban_perm);
	adc_msg_free(hub->command_ban_temp);
	adc_msg_free(hub->command_registered_only);
	adc_msg_free(hub->command_password_wrong);
	adc_msg_free(hub->command_error_hub_full);
	adc_msg_free(hub->command_error_hub_disabled);
	adc_msg_free(hub->command_ping);
}



int hub_handle_info_check_nick(struct hub_info* hub, struct user* user, struct adc_message* cmd)
{
	size_t length, pos;
	int nicks;
	char* arg;
	struct user* lookup;
	
	nicks = adc_msg_has_named_argument(cmd, "NI");
	if (nicks != 1)
	{
		if (nicks == 0)
			on_login_failure(hub, user, hub->command_error_nick_missing, "nickname missing");
		else
			on_login_failure(hub, user, hub->command_error_nick_invalid, "multiple nicknames given");
		return 0;
	}
	
	arg = adc_msg_get_named_argument(cmd, "NI");
	if (arg == NULL)
		return 0;
	
	length = strlen(arg);

	/* Length of nick must at least be 2 characters long */
	if (length <= 1 || length > MAX_NICK_LEN)
	{
		on_login_failure(hub, user, hub->command_error_nick_invalid, "nickname has invalid length");
		hub_free(arg);
		return 0;
	}
	
	/* Nick must not contain newline */
	for (pos = 1; pos < length; pos++)
	{
		if (arg[pos-1] == '\\' && arg[pos] == 'n')
		{
			route_to_user(user, hub->command_error_nick_invalid);
			on_login_failure(hub, user, hub->command_error_nick_invalid, "nickname contains newline");
			hub_free(arg);
			return 0;
		}
	}

	/* Nick must not start with a space */
	if (arg[0] == '\\' && arg[1] == 's') /* FIXME: Check for other types of spaces */
	{
		on_login_failure(hub, user, hub->command_error_nick_invalid, "nickname starts with space");
		hub_free(arg);
		return 0;
	}
	
	/*
	 * Nick should be valid utf-8, but
	 * perhaps we should check if the nick is unicode normalized?
	 */
	if (!is_valid_utf8(arg))
	{
		on_login_failure(hub, user, hub->command_error_nick_invalid, "nickname is not valid utf-8");
		hub_free(arg);
		return 0;
	}
	
	/*
	 * Since the hub sometimes send messages from itself,
	 * we do not want users named the same as the hub.
	 */
	if (!strcmp(arg, hub->config->hub_name))
	{
		on_login_failure(hub, user, hub->command_error_nick_invalid, "nickname is the same as hub name");
		hub_free(arg);
		return 0;
	}
	
	lookup = get_user_by_nick(hub, arg);
	if (lookup && lookup->sid != user->sid)
	{
		on_login_failure(hub, user, hub->command_error_nick_invalid, "nickname is already taken");
		hub_free(arg);
		return 0;
	}
	else
	{
		/* hub_log(log_trace, "User '%s' changed nickname to '%s'", user->nick, arg); */
		
		/* Set nickname */
		memcpy(user->nick, arg, length);
		user->nick[length] = 0;
	}
	
	hub_free(arg);
	return 1;
}


int hub_handle_info_check_cid(struct hub_info* hub, struct user* user, struct adc_message* cmd)
{
	size_t pos;
	char* cid = adc_msg_get_named_argument(cmd, "ID");
	char* pid = adc_msg_get_named_argument(cmd, "PD");
	char x_pid[64];
	char raw_pid[64];
	uint64_t tiger_res[3];
	struct user* lookup = NULL;

	if (cid == NULL || pid == NULL)
	{
		hub_log(log_debug, "adc_inf_check_cid(): no cid (%p) or pid (%p)", cid, pid);
		
		if (cid == NULL)
			route_to_user(user, hub->command_error_cid_missing);
		else if (pid == NULL)
			route_to_user(user, hub->command_error_pid_missing);
		
		hub_free(cid);
		hub_free(pid);
		return 0;
	}
	
	if (strlen(cid) != strlen(pid) || strlen(cid) != MAX_CID_LEN)
	{
		route_to_user(user, hub->command_error_cid_invalid);
		hub_free(cid);
		hub_free(pid);
		return 0;
	}

	for (pos = 0; pos < MAX_CID_LEN; pos++)
	{
		if (!is_valid_base32_char(cid[pos]) || !is_valid_base32_char(pid[pos]))
		{
			route_to_user(user, hub->command_error_cid_invalid);
			hub_free(cid);
			hub_free(pid);
			return 0;
		}
	}
	
	memset(x_pid, 0, MAX_CID_LEN+1);
	base32_decode(pid, (unsigned char*) raw_pid, MAX_CID_LEN);
	tiger((uint64_t*) raw_pid, TIGERSIZE, (uint64_t*) tiger_res);
	base32_encode((unsigned char*) tiger_res, TIGERSIZE, x_pid);
	
	x_pid[MAX_CID_LEN] = 0;
	
	hub_free(pid); /* We don't need it past here */
	
	if (strncasecmp(x_pid, cid, MAX_CID_LEN) != 0)
	{
		hub_log(log_trace, "cid/pid mismatch");
		route_to_user(user, hub->command_error_cid_invalid);
		hub_free(cid);
		return 0;
	}
	
	lookup = get_user_by_cid(hub, cid);
	if (lookup && lookup->sid != user->sid)
	{
		route_to_user(user, hub->command_error_cid_taken);
		hub_free(cid);
		return 0;
	}
	else
	{
		/* Set cid */
		memcpy(user->cid, cid, MAX_CID_LEN);
		user->cid[MAX_CID_LEN] = 0;
	}

	hub_free(cid);
	return 1;
}


