/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "uhub.h"

static int verbosity = 4;
static FILE* logfile = NULL;

#ifdef MEMORY_DEBUG
static FILE* memfile = NULL;
#define MEMORY_DEBUG_FILE "memlog.txt"
#endif

static const char* prefixes[] =
{
	"FATAL",
	"ERROR",
	"WARN",
	"USER",
	"INFO",
	"DEBUG",
	"TRACE",
	"DUMP",
    "MEM",
	0
};


void hub_log_initialize(const char* file)
{
#ifdef MEMORY_DEBUG
	memfile = fopen(MEMORY_DEBUG_FILE, "w");
	if (!memfile)
	{
		fprintf(stderr, "Unable to create " MEMORY_DEBUG_FILE " for logging memory allocations\n");
		return;
	}
#endif
	
	if (!file)
	{
		logfile = stderr;
		return;
	}
	
	logfile = fopen(file, "a");
	if (!logfile)
	{
		logfile = stderr;
		return;
	}
}


void hub_log_shutdown()
{
	if (logfile && logfile != stderr)
	{
		fclose(logfile);
		logfile = NULL;
	}

#ifdef MEMORY_DEBUG
	if (memfile)
	{
		fclose(memfile);
		memfile = NULL;
	}
#endif
}


void hub_set_log_verbosity(int verb)
{
	verbosity = verb;
}

void hub_log(int log_verbosity, const char *format, ...)
{
	static char logmsg[1024];
	static char timestamp[32];
	struct tm *tmp;
	time_t t;
	va_list args;

#ifdef MEMORY_DEBUG
	if (memfile && log_verbosity == log_memory)
	{
		va_start(args, format);
		vsnprintf(logmsg, 1024, format, args);
		va_end(args);
		fprintf(memfile, "%s\n", logmsg);
		fflush(memfile);
		return;
	}
#endif
	
	if (log_verbosity < verbosity && logfile)
	{
		t = time(NULL);
		tmp = localtime(&t);
		strftime(timestamp, 32, "%a %e %T", tmp);
		va_start(args, format);
		vsnprintf(logmsg, 1024, format, args);
		va_end(args);
		fprintf(logfile, "%s %5s: %s\n", timestamp, prefixes[log_verbosity], logmsg);
		fflush(logfile);
	}
}
