/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "uhub.h"

#ifdef DEBUG
#define ADC_MSG_ASSERT(X) \
	uhub_assert(X); \
	uhub_assert(X->cache); \
	uhub_assert(X->capacity); \
	uhub_assert(X->length); \
	uhub_assert(X->length <= X->capacity); \
	uhub_assert(X->arg_offset < X->length); \
	uhub_assert(X->length == strlen(X->cache)); \
	uhub_assert(X->references >= 0);
#else
#define ADC_MSG_ASSERT(X) do { } while(0);
#endif


// FIXME: This is a temporary solution!
struct adc_message* adc_msg_incref(struct adc_message* msg)
{
#ifndef ADC_MESSAGE_INCREF
	msg->references++;
	return msg;
#else
	struct adc_message* copy = adc_msg_copy(msg);
	return copy;
#endif
}


void adc_msg_free(struct adc_message* msg)
{
	if (!msg) return;
	
	ADC_MSG_ASSERT(msg);
	
	if (msg->references > 0)
	{
		msg->references--;
	}
	else
	{
		hub_free(msg->cache);
		
		if (msg->feature_cast_include)
		{
			list_clear(msg->feature_cast_include, &hub_free);
			list_destroy(msg->feature_cast_include);
			msg->feature_cast_include = 0;
		}
		
		if (msg->feature_cast_exclude)
		{
			list_clear(msg->feature_cast_exclude, &hub_free);
			list_destroy(msg->feature_cast_exclude);
			msg->feature_cast_exclude = 0;
		}
		
		hub_free(msg);
	}
}


struct adc_message* adc_msg_copy(struct adc_message* cmd)
{
	char* tmp = 0;
	struct adc_message* copy = (struct adc_message*) hub_malloc_zero(sizeof(struct adc_message));
	if (!copy) return NULL; /* OOM */

	ADC_MSG_ASSERT(cmd);

	memcpy(copy, cmd, sizeof(struct adc_message));
	if (cmd->cache)
	{
		copy->cache = hub_malloc(cmd->length + 2);
		memcpy(copy->cache, cmd->cache, cmd->length);
		copy->cache[cmd->length] = 0;
	}
	
	if (cmd->feature_cast_include)
	{
		copy->feature_cast_include = list_create();
		tmp = list_get_first(cmd->feature_cast_include);
		while (tmp)
		{
			list_append(copy->feature_cast_include, hub_strdup(tmp));
			tmp = list_get_next(cmd->feature_cast_include);
		}
	}
	
	if (cmd->feature_cast_exclude)
	{
		copy->feature_cast_exclude = list_create();
		tmp = list_get_first(cmd->feature_cast_exclude);
		while (tmp)
		{
			list_append(copy->feature_cast_exclude, hub_strdup(tmp));
			tmp = list_get_next(cmd->feature_cast_exclude);
		}
	}
	
	ADC_MSG_ASSERT(copy);
	
	return copy;
}


struct adc_message* adc_msg_parse(struct user* u, const char* line, size_t length)
{
	struct adc_message* command = (struct adc_message*) hub_malloc_zero(sizeof(struct adc_message));
	char prefix = line[0];
	size_t n = 0;
	char temp_sid[5];
	int ok = 1;
	int need_terminate = 0;
	struct linked_list* feature_cast_list;
	
	if (command == NULL)
		return NULL; /* OOM */

	if (line[length-1] != '\n')
	{
		need_terminate = 1;
	}

	command->length = length + need_terminate;
	command->capacity = command->length;
	command->cache = (char*) hub_malloc(command->capacity+1);
	
	if (command->cache == NULL)
	{
		hub_free(command);
		return NULL; /* OOM */
	}
	
	memcpy(command->cache, line, length);
	
	/* Ensure we are zero terminated */
	command->cache[length] = 0;
	command->cache[length+need_terminate] = 0;
	
	command->cmd = FOURCC(line[0], line[1], line[2], line[3]);
	command->priority = 0;
	command->arg_offset = 4;
	
	switch (prefix)
	{
		case 'U':
		case 'C':
			/* these should never be seen on a hub */
			ok = 0;
			break;
			
		case 'I':
		case 'H':
			ok = (length > 3);
			command->arg_offset = 4;
			break;
			
		case 'B':
			if (!u) ok = 0;
			if (!ok) break;
			
			ok = (length > 8 &&
					is_space(line[4]) &&
					is_valid_base32_char(line[5]) &&
					is_valid_base32_char(line[6]) &&
					is_valid_base32_char(line[7]) &&
					is_valid_base32_char(line[8]));
			
			if (!ok) break;
			
			temp_sid[0] = line[5];
			temp_sid[1] = line[6];
			temp_sid[2] = line[7];
			temp_sid[3] = line[8];
			temp_sid[4] = '\0';
			
			command->source = string_to_sid(temp_sid);
			ok = (command->source == u->sid);
			
			command->arg_offset = 9;
			break;
			
		case 'F':
			if (!u) ok = 0;
			if (!ok) break;
			
			ok = (length > 8 &&
					is_space(line[4]) &&
					is_valid_base32_char(line[5]) &&
					is_valid_base32_char(line[6]) &&
					is_valid_base32_char(line[7]) &&
					is_valid_base32_char(line[8]));
			
			if (!ok) break;
			
			temp_sid[0] = line[5];
			temp_sid[1] = line[6];
			temp_sid[2] = line[7];
			temp_sid[3] = line[8];
			temp_sid[4] = '\0';
			
			command->source = string_to_sid(temp_sid);
			ok = (command->source == u->sid);
			
			/* Create feature cast lists */
			command->feature_cast_include = list_create();
			command->feature_cast_exclude = list_create();
			
			if (!command->feature_cast_include || !command->feature_cast_exclude)
			{
				list_destroy(command->feature_cast_include);
				list_destroy(command->feature_cast_exclude);
				hub_free(command->cache);
				hub_free(command);
				return NULL; /* OOM */
			}
			
			n = 10;
			while (line[n] == '+' || line[n] == '-')
			{
				if (line[n++] == '+')
					feature_cast_list = command->feature_cast_include;
				else
					feature_cast_list = command->feature_cast_exclude;
				
				temp_sid[0] = line[n++];
				temp_sid[1] = line[n++];
				temp_sid[2] = line[n++];
				temp_sid[3] = line[n++];
				temp_sid[4] = '\0';
				
				list_append(feature_cast_list, hub_strdup(temp_sid));
			}
			
			command->arg_offset = n;
			
			if  (n == 10)
				ok = 0;
			
			break;
			
		case 'D':
		case 'E':
			if (!u) ok = 0;
			if (!ok) break;
			
			ok = (length > 13 &&
					is_space(line[4]) &&
					is_valid_base32_char(line[5]) &&
					is_valid_base32_char(line[6]) &&
					is_valid_base32_char(line[7]) &&
					is_valid_base32_char(line[8]) &&
					is_space(line[9]) &&
					is_valid_base32_char(line[10]) &&
					is_valid_base32_char(line[11]) &&
					is_valid_base32_char(line[12]) &&
					is_valid_base32_char(line[13]));
			
			if (!ok) break;
			
			temp_sid[0] = line[5];
			temp_sid[1] = line[6];
			temp_sid[2] = line[7];
			temp_sid[3] = line[8];
			temp_sid[4] = '\0';
			
			command->source = string_to_sid(temp_sid);
			ok = (command->source == u->sid);
			
			if (!ok) break;
			
			temp_sid[0] = line[10];
			temp_sid[1] = line[11];
			temp_sid[2] = line[12];
			temp_sid[3] = line[13];
			temp_sid[4] = '\0';
			
			command->target = string_to_sid(temp_sid);
			
			command->arg_offset = 14;
			break;
			
		default:
			ok = 0;
	}
	
	if (need_terminate)
	{
		command->cache[length] = '\n';
	}
	
	if (!ok) {
		adc_msg_free(command);
		return NULL;
	}
	
	/* At this point the arg_offset should point to a space, or the end of message */
	if (command->cache[command->arg_offset] == ' ')
	{
		command->arg_offset++;
		if (command->cache[command->arg_offset] == ' ') ok = 0;
	}
	else if (command->cache[command->arg_offset] == '\n') ok = 1;
	else ok = 0;
	
	if (!ok)
	{
		adc_msg_free(command);
		return NULL;
	}
	
	ADC_MSG_ASSERT(command);
	
	return command;
}


struct adc_message* adc_msg_create(const char* line)
{
	return adc_msg_parse(NULL, line, strlen(line));
}

int adc_msg_remove_named_argument(struct adc_message* cmd, const char prefix_[2])
{
	char* start;
	char* end;
	char* endInfo;
	size_t endlen;
	char prefix[4] = { ' ', prefix_[0], prefix_[1], '\0' };
	int found = 0;
	size_t temp_len = 0;
	
	adc_msg_unterminate_command(cmd);
	
	start = memmem(&cmd->cache[cmd->arg_offset-1], (cmd->length - cmd->arg_offset), prefix, 3);
	while (start)
	{
		endInfo = &cmd->cache[cmd->length];
	
		if  (&start[0] < &endInfo[0])
		{
			end = memchr(&start[1], ' ', &endInfo[0]-&start[1]);
		}
		else
		{
			end = NULL;
		}
		
		if (end)
		{
			
			temp_len = &end[0] - &start[0]; // strlen(start);
			/* hub_log(log_trace, "                     length=%d", (int) (temp_len)); */
			endlen = strlen(end);
			memmove(start, end, endlen);
			start[endlen] = '\0';
			found++;
			cmd->length -= temp_len;
		}
		else
		{
			found++;
			cmd->length -= strlen(start);
			start[0] = '\0';
			break;
		}
		start = memmem(&cmd->cache[cmd->arg_offset-1], (cmd->length - cmd->arg_offset), prefix, 3);
	}
	
	adc_msg_terminate_command(cmd);
	
	return found;
}


int adc_msg_has_named_argument(struct adc_message* cmd, const char prefix_[2])
{
	int count = 0;
	char* start;
	char prefix[4] = { ' ', prefix_[0], prefix_[1], '\0' };

	ADC_MSG_ASSERT(cmd);
	
	start = memmem(&cmd->cache[cmd->arg_offset-1], (cmd->length - cmd->arg_offset), prefix, 3);
	while (start)
	{
		count++;
		if ((&start[0] - &cmd->cache[0])  > 1+cmd->length)
			start = memmem(&start[1], (&cmd->cache[cmd->length] - &start[0]), prefix, 3);
		else
			start = NULL;
	}
	
	return count;
}


char* adc_msg_get_named_argument(struct adc_message* cmd, const char prefix_[2])
{
	char* start;
	char* end;
	char* argument;
	size_t length;
	char prefix[4] = { ' ', prefix_[0], prefix_[1], '\0' };
	
	ADC_MSG_ASSERT(cmd);
	
	start = memmem(&cmd->cache[cmd->arg_offset-1], cmd->length - cmd->arg_offset, prefix, 3);
	if (!start)
		return NULL;
	
	start = &start[3];
	end = strchr(start, ' ');
	if (!end) end = &cmd->cache[cmd->length];
	length = &end[0] - &start[0];
	
	argument = hub_malloc(length + 1);
	if (!argument)
		return NULL;

	memcpy(argument, start, length);
	argument[length] = 0;
	
	if (length > 0 && argument[length-1] == '\n')
	{
		argument[length-1] = 0;
	}
	
	return argument;
}


int adc_msg_replace_named_argument(struct adc_message* cmd, const char prefix[2], const char* string)
{
	ADC_MSG_ASSERT(cmd);
	
	if (adc_msg_has_named_argument(cmd, prefix))
	{
		adc_msg_remove_named_argument(cmd, prefix);
	}
	
	if (adc_msg_add_named_argument(cmd, prefix, string) == -1)
	{
		return -1;
	}
	
	ADC_MSG_ASSERT(cmd);
	
	return 0;
}


void adc_msg_terminate_command(struct adc_message* cmd)
{
	char* data = cmd->cache;
	size_t len = cmd->length;
	
	uhub_assert(cmd->cache[cmd->length -1] != '\n');
	
	if (len > 0 && data[len-1] != '\n')
	{
		if (len >= cmd->capacity)
		{
			data = hub_malloc(len + 2); /* FIXME: OOM! */
			memcpy(data, cmd->cache, cmd->length);
			hub_free(cmd->cache);
			cmd->cache = data;
			cmd->capacity = len+2;
		}
		
		data[cmd->length++] = '\n';
		data[cmd->length] = 0;
	}
	
	ADC_MSG_ASSERT(cmd);
}


void adc_msg_unterminate_command(struct adc_message* cmd)
{
	ADC_MSG_ASSERT(cmd);
	
	uhub_assert(cmd->cache[cmd->length -1] == '\n');
	
	if (cmd->length > 0 && cmd->cache[cmd->length-1] == '\n')
	{
		cmd->length--;
		cmd->cache[cmd->length] = 0;
	}
}

int adc_msg_add_named_argument(struct adc_message* cmd, const char prefix_[2], const char* string)
{
	int ret = 0;
	char prefix[3] = { prefix_[0], prefix_[1], 0 };
	size_t length = strlen(string);
	size_t add_length = length + 2; // len + prefix
	char* str = hub_malloc(add_length+1);
	
	if (!str) return -1;
	
	ADC_MSG_ASSERT(cmd);
	
	memcpy(str, prefix, 2);
	memcpy(&str[2], string, length);
	str[add_length] = 0;
	
	ret = adc_msg_add_argument(cmd, str);
	hub_free(str);
	return ret;
}


int adc_msg_add_argument(struct adc_message* cmd, const char* string)
{
	char* data = 0;
	size_t len = 0;
	size_t capacity = 0;
	size_t add_length = strlen(string) + 1; /* including command separator (space) */
	/* hub_log(log_trace, "Add argument: '%s' (%d bytes)", string, (int) add_length); */
	
	ADC_MSG_ASSERT(cmd);
	
	adc_msg_unterminate_command(cmd);
	
	len = cmd->length + add_length;
	if (len >= cmd->capacity)
	{
		capacity = len + 1;
		data = hub_malloc(capacity+1);
		if (data == NULL)
		{
			adc_msg_terminate_command(cmd);
			return -1;
		}
		memcpy(data, cmd->cache, cmd->length);
		
		hub_free(cmd->cache);
		cmd->cache = data;
		cmd->capacity = capacity;
	}
	else
	{
		data = cmd->cache;
	}
	
	data[cmd->length] = ' ';
	memcpy(&data[cmd->length+1], string, strlen(string));
	cmd->length += add_length;
	
	adc_msg_terminate_command(cmd);
	return 0;
}


char* adc_msg_get_argument(struct adc_message* cmd, int offset)
{
	char* start;
	char* end;
	char* argument;
	int count = 0;
	
	ADC_MSG_ASSERT(cmd);
	
	adc_msg_unterminate_command(cmd);
	
	start = strchr(&cmd->cache[cmd->arg_offset-1], ' ');
	while (start)
	{
		end = strchr(&start[1], ' ');
		
		if (count == offset)
		{
			if (end)
			{
				argument = hub_strndup(&start[1], (&end[0] - &start[1]));
			}
			else
			{
				argument = hub_strdup(&start[1]);
				if (argument[strlen(argument)-1] == '\n')
					argument[strlen(argument)-1] = 0;
			}
			
			if (strlen(argument))
			{
				adc_msg_terminate_command(cmd);
				return argument;
			}
		}
		
		count++;
		start = end;
	}
	
	adc_msg_terminate_command(cmd);
	return 0;
}


int adc_msg_escape_length(const char* str)
{
	int add = 0;
	int n = 0;
	for (; str[n]; n++)
		if (str[n] == ' ' || str[n] == '\n' || str[n] == '\\') add++;
	return n + add;
}


int adc_msg_unescape_length(const char* str)
{
	int add = 0;
	int n = 0;
	int escape = 0;
	for (; str[n]; n++)
	{
		if (escape)
		{
			escape = 0;
		}
		else
		{
			if (str[n] == '\\')
			{
				escape = 1;
				add++;
			}
		}
	}
	return n - add;
}


char* adc_msg_unescape(const char* string)
{
	char* new_string = hub_malloc(adc_msg_unescape_length(string)+1);
	char* ptr = (char*) new_string;
	char* str = (char*) string;
	int escaped = 0;
	
	while (*str)
	{
		if (escaped) {
			if (*str == 's')
				*ptr++ = ' ';
			else if (*str == '\\')
				*ptr++ = '\\';
			else if (*str == 'n')
				*ptr++ = '\n';
			else
				*ptr++ = *str;
			
			escaped = 0;
		} else {
			if (*str == '\\')
				escaped = 1;
			else
				*ptr++ = *str;
				
		}
		str++;
	}
	*ptr = 0;
	return new_string;
}


char* adc_msg_escape(const char* string)
{
	char* str = hub_malloc(adc_msg_escape_length(string)+1);
	int n = 0;
	int i = 0;
	for (i = 0; i < strlen(string); i++)
	{
		switch (string[i]) {
			case '\\': /* fall through */
					str[n++] = '\\';
					str[n++] = '\\';
					break;
			case '\n':
					str[n++] = '\\';
					str[n++] = 'n';
					break;
			case ' ':
					str[n++] = '\\';
					str[n++] = 's';
					break;
			default:
					str[n++] = string[i];
					break;
		}
	}
	str[n] = '\0';
	return str;
}

