/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "uhub.h"


/*
 * This callback function is used to clear user objects from the userlist.
 * Should only be used in user_manager_shutdown().
 */
static void clear_user_list_callback(void* ptr)
{
	if (ptr)
	{
		struct user* u = (struct user*) ptr;
		
		/* Mark the user as already being disconnected.
		 * This prevents the hub from trying to send
		 * quit messages to other users.
		 */
		u->credentials = cred_none;
		user_destroy(u);
	}
}

static void timer_statistics(int fd, short ev, void *arg)
{
	struct hub_info* hub = (struct hub_info*) arg;
	struct timeval timeout = { TIMEOUT_STATS, 0 };
	int factor = 0;
	
	struct net_statistics* total;
	struct net_statistics* intermediate;
		
	net_stats_get(&intermediate, &total);
	
	factor = (time(NULL) - intermediate->timestamp);
	if (!factor) factor++;

	hub_log(log_info, "Statistics  users=%d, net_tx=%d KB/s, net_rx=%d KB/s",
		(int) list_size(hub->user_list),
		(int) ((intermediate->tx / factor) / 1024),
		(int) ((intermediate->rx / factor) / 1024));
	
	net_stats_reset();
	
	evtimer_set(&hub->ev_timer, timer_statistics, hub);
	evtimer_add(&hub->ev_timer, &timeout);
}



int user_manager_init(struct hub_info* hub)
{
	struct timeval timeout = { TIMEOUT_STATS, 0 };

	hub->user_list = list_create();
	
	if (!hub->user_list)
		return -1;
	
	evtimer_set(&hub->ev_timer, timer_statistics, hub);
	evtimer_add(&hub->ev_timer, &timeout);
	return 0;
}


void user_manager_shutdown(struct hub_info* hub)
{
	event_del(&hub->ev_timer);
	list_clear(hub->user_list, &clear_user_list_callback);
	list_destroy(hub->user_list);
}




struct user* user_create(struct hub_info* hub, int sd)
{
	struct user* user = NULL;

	user = (struct user*) hub_malloc_zero(sizeof(struct user));

	if (user == NULL)
		return NULL; /* OOM */
	
	user->sd = sd;
	user->sid = get_free_sid(hub);
	user->tm_connected = time(NULL);
	user->hub = hub;
	user->feature_cast = 0;
	
	user->send_queue = list_create();
	user->send_queue_offset = 0;
	user->send_queue_size = 0;
	
	user->recv_buf_offset = 0;
	user->recv_buf = 0;

	list_append(hub->user_list, user);

	hub_log(log_debug, "user_create(): adding new user (sid=%d, sd=%d)", (int) user->sid, (int) user->sd);
	return user;
}

static void clear_send_queue_callback(void* ptr)
{
	adc_msg_free((struct adc_message*) ptr);
}

void user_destroy(struct user* user)
{
	if (user->credentials != cred_none && user->hub->status == hub_status_running)
	{
		user->state = state_disconnected;
		send_quit_message(user->hub, user);
	}
	
	if (user->hub->status == hub_status_running)
	{
		list_remove(user->hub->user_list, user);
	}
	
	on_logout_user(user->hub, user);

	adc_msg_free(user->info);
	
	if (user->feature_cast)
	{
		list_clear(user->feature_cast, &hub_free);
		list_destroy(user->feature_cast);
		user->feature_cast = 0;
	}
	
	if (user->recv_buf)
	{
		hub_free(user->recv_buf);
		user->recv_buf = 0;
		user->recv_buf_offset = 0;
	}
	
	hub_log(log_debug, "user_destroy(): removing user (sid=%d, sd=%d, send_queue=%d (%d bytes))", (int) user->sid, (int) user->sd, (int) (user->send_queue ? user->send_queue->size : 0), (int) user->send_queue_size);
	
	if (user->send_queue)
	{
		list_clear(user->send_queue, &clear_send_queue_callback);
		list_destroy(user->send_queue);
		user->send_queue = 0;
	}
	
	event_del(&user->ev_write);
	event_del(&user->ev_read);
	net_close(user->sd);
	hub_free(user);
}


void user_set_info(struct user* user, struct adc_message* info)
{
	adc_msg_free(user->info);
	user->info = info;
}


static int convert_support_fourcc(int fourcc)
{
	switch (fourcc)
	{
		case FOURCC('B','A','S','0'):
		case FOURCC('B','A','S','E'):
			return feature_base;
			
		case FOURCC('A','U','T','0'):
			return  feature_auto;
			
		case FOURCC('U','C','M','D'):
			return feature_ucmd;
			
		case FOURCC('Z','L','I','F'):
			return feature_zlif;
			
		case FOURCC('B','B','S','0'):
			return feature_bbs;
			
		default:
			hub_log(log_debug, "Unknown extension: %x", fourcc);
			return 0;
	}
}


void user_support_add(struct user* user, int fourcc)
{
	int feature_mask = convert_support_fourcc(fourcc);
	user->features |= feature_mask;
}


void user_support_remove(struct user* user, int fourcc)
{
	int feature_mask = convert_support_fourcc(fourcc);
	user->features &= ~feature_mask;
}


struct user* get_user_by_sid(struct hub_info* hub, sid_t sid)
{
	struct user* user = (struct user*) list_get_first(hub->user_list);
	while (user)
	{
		if (user->sid == sid)
			return user;
		user = (struct user*) list_get_next(hub->user_list);
	}
	return NULL;
}


struct user* get_user_by_nick(struct hub_info* hub, const char* nick)
{
	struct user* user = (struct user*) list_get_first(hub->user_list);
	while (user)
	{
		if (user->state == state_verify || user->state == state_normal)
		{
			if (strcmp(user->nick, nick) == 0)
				return user;
		}
		user = (struct user*) list_get_next(hub->user_list);
	}
	return NULL;
}


struct user* get_user_by_cid(struct hub_info* hub, const char* cid)
{
	struct user* user = (struct user*) list_get_first(hub->user_list);
	while (user)
	{
		if (user->state == state_verify || user->state == state_normal)
		{
			if (strcmp(user->cid, cid) == 0)
				return user;
		}
		user = (struct user*) list_get_next(hub->user_list);
	}
	return NULL;
}


void send_queued_data(struct user* user)
{

}


void user_disconnect(struct user* user, int reason)
{
	struct timeval timeout = { 0, 0 };
	event_del(&user->ev_write);
	event_del(&user->ev_read);
	user->state = state_wait_disconnect;
	user->quit_reason = reason;
	event_add(&user->ev_write, &timeout);
}


int user_have_feature_cast_support(struct user* user, char feature[4])
{
	char* tmp = list_get_first(user->feature_cast);
	while (tmp)
	{
		if (strncmp(tmp, feature, 4) == 0)
			return 1;
	
		tmp = list_get_next(user->feature_cast);
	}
	
	return 0;
}


void send_user_list(struct hub_info* hub, struct user* target)
{
	struct user* user = (struct user*) list_get_first(hub->user_list);
	while (user)
	{
		if (user->state == state_normal)
		{
			route_to_user(target, user->info);
		}
		user = (struct user*) list_get_next(hub->user_list);
	}
}


void send_quit_message(struct hub_info* hub, struct user* leaving)
{
	struct adc_message* command = adc_msg_create("IQUI");
	adc_msg_add_argument(command, (const char*) sid_to_string(leaving->sid));
	
	if (leaving->quit_reason == quit_banned || leaving->quit_reason == quit_kicked)
	{
		adc_msg_add_argument(command, (const char*) "DI");
	}
	
	route_to_all(hub, command);
	adc_msg_free(command);
}


